## Why are Plutus Core validators so compressible?

As we saw in [Erasure.md](./Erasure.md), we can reduce the size of CBOR-serialised Plutus Core programs by performing two types of transformation:

   * "Minimising" ASTs by removing superpfluous information:
        * Remove types and type-related instructions
        * Remove textual identifiers from names
        * Replace names with de Bruijn indices (at least for transmission)
        * Remove annotations from all AST nodes during serialisation (these are all `()` most of the time anyway)
   * Compressing the minimised AST.

Minimisation reduces the size or serialised ASTs by a factor of about 15, and compression reduces the size by a further factor of 3 of 4 (using `gzip`; you can get somewhat better results using `bzip2` or `xz`), for an overall size reduction by factor of roughly 50 to 60.

The question arises of why it is that serialised ASTs are so compressible.  A close inspection of the CBOR reveals that they are mostly empty space.  The type-erased AST has only six constructors (variables, lambda abstractions, applications, values of built-in types (called `Constant`s), names of built-in functions, and the `Error` term), and these are serialised into the integers 0-5.  These, together with de Bruijn indices, constitute the majority of the AST, and the result is that the CBOR is rather sparse.  I counted the number of bits set in each of our serialised validators, and it turns out that on average only about one bit per byte is actually set: see the table below and the dump of the Crowdfunding validator at the end of this file.

| Contract | Bytes in miminised serialised validator | Nonzero bits | Nonzero bits/byte | Compressed size |
| :---                 |  ---: |  ---: | ---: | ---: |
| Crowdfunding         |  6593 |  6156 | 0.93 | 1541 |
| Currrency            | 10655 | 11469 | 1.08 | 3010 |
| Escrow               |  7653 |  8053 | 1.05 | 1882 |
| Future               | 13914 | 15829 | 1.14 | 3394 |
| Game                 |  6635 |  6990 | 1.05 | 2148 |
| GameStateMachine     |  8997 | 11065 | 1.23 | 1980 |
| MultiSig             |  8149 |  8897 | 1.09 | 2552 |
| MultiSigStateMachine | 14956 | 17270 | 1.15 | 3719 |
| PubKey               |  7579 |  8143 | 1.07 | 2409 |
| Swap                 | 12077 | 12838 | 1.06 | 3756 |
| TokenAccount         |  3184 |  2477 | 0.78 |  659 |
| Vesting              |  6967 |  6480 | 0.93 | 1614 |

Not only is the representation very sparse, but the distribution of bytes is highly skewed.  For the Crowdfunding validator, the frequency of bytes is as follows:

| Byte | Number of occurrences |
| :--- | --------------------: |
| 0    |	2393	       |
| 2    |	1435	       |
| 1    |	1397	       |
| 24   |	392	       |
| 3    |	115	       |
| 5    |	104	       |
| 4    |	101	       |
| 138  |	39	       |
| 6    |	37	       |
| 7    |	28	       |
| 9    |	21	       |
| 64   |	18	       |
| 65   |	17	       |
| 14   |	16	       |
| 8, 10, 11, 13, 15, 63 |  15  |
| 12, 47, 48   | 14	       |
| 16, 17, 32, 49, 66, 105   | 10 |
| 22, 23, 43, 67, 110  | 8     |
| 34, 46, 50, 62, 96   | 7     |
| 18, 26, 37, 42, 44, 116  | 6 |

We also have (but it doesn't fit in the table)
   * 5 occurrences: 20, 31, 35, 45, 61, 101, 107, 108, 114, 120
   * 4 occurrences: 19, 27, 28, 33, 36, 38, 39, 41, 68, 104, 112, 117, 121
   * 3 occurrences: 21, 25, 29, 40, 60, 99, 103, 109, 115, 128
   * 2 occurrences: 59, 84, 85, 100, 119, 127, 130,  41
   * 1 occurrence:  30, 52, 55, 58, 69, 81, 83, 90, 91, 95, 96, 98, 102, 106, 111, 118, 122, 124, 125, 126, 129, 135, 139, 140, 142

Given this, it's not surprising that the data is highly compressible.  Typical compression algorithms do two main things: (1) replace repeated substrings with a reference to a single copy, and (2) use some variable-width encoding such as Huffman encoding to encode more frequent input bytes with a smaller number of bits. The skewed distribution of the inputs bytes here helps with this: for example, a straightforward Huffman encoding with no attempt to remove repetitions reduces the size of the Crowdfunding validator from 6593 bytes to 2551 bytes.

**Note.** We might want to think about how this interacts with
Duncan's proposal to extend CBOR to handle datatypes.  Here we're
just emitting a single CBOR byte for each constructor in the AST
term representation.

### Consequences
The sparsity of serialised validators suggests two things:
   * There's probably not too much point in trying to reduce redundancy in the AST (by sharing terms, for example).  There doesn't in fact seem to be too much repetition, and compression will take care of repetition for us anyway.
   * If we really want to make sizes as small as possible for transmission, we probably _have_ to use compression.  We have quite a lot of bytes but not much information per byte, so we have to squash them together in some way.
   
#### Merklisation
This may also tell us something about why Merklisation doesn't seem to be very effective.  A Merkle hash consists of 32 effectively random bytes and will probably be pretty much incompressible. Moreover, introducing lots of random bytes might even reduce the compressibility of the remaining code by reducing repetitions and disrupting the frequencies used to construct Huffman encodings.

You'd expect about 50% of the bits in a hash to be set, so in total a 32-byte hash will contain about 32 * (8/2) = 128 bits, about the same number of bits as 128 _bytes_ of our uncompressed serialised AST.  This suggests that you'd really need to be Merklising away 128 bytes just to break even, and more than 128 bytes to see a saving. [This is all just handwaving: I should learn something about information theory.]

### Appendix: the Crowdfunding validator

For completeness, this section includes a dump of the serialised minimised validator for the Crowdfunding example, obtained with `od -t u1`.  Each line contains the decimal contents of 16 bytes, and the numbers down the left hand side are hexadecimal offsets.  There are a total of 6593 bytes.  If you scan through it, you'll see that it's very very sparse.

The validator starts with a long chain of lambda applications: `[[ (lam 0 (lam 0 [ [ [ (lam 0 (lam 0 (lam 0 [ (lam 0 [ (lam 0 [ [ (lam 0 (lam 0 [ (lam 0 [ [ [ [ (lam 0 (lam 0 (lam 0 (lam 0 [ [ [ (lam 0 (lam 0 (lam 0 [ (lam 0 [ ...`  This accounts for all of the zeros, ones, and twos at the start.

```
0000000   1   0   0   2   2   1   0   1   0   2   2   2   1   0   1   0      First three bytes are PLC version (1.0.0)
0000020   1   0   2   1   0   2   1   0   2   2   1   0   1   0   2   1
0000040   0   2   2   2   2   1   0   1   0   1   0   1   0   2   2   2
0000060   1   0   1   0   1   0   2   1   0   2   1   0   2   1   0   2
0000100   1   0   2   1   0   2   1   0   2   1   0   2   2   2   2   1      CBOR encodings for AST term constructors:
0000120   0   1   0   1   0   1   0   2   1   0   2   1   0   2   1   0	                                              
0000140   2   1   0   2   1   0   2   2   1   0   1   0   2   1   0   2         0   Var				      
0000160   1   0   2   1   0   2   1   0   2   2   2   2   1   0   1   0         1   LamAbs			      
0000200   1   0   1   0   2   1   0   2   1   0   2   1   0   2   2   1         2   Apply			      
0000220   0   1   0   2   1   0   2   1   0   2   2   1   0   1   0   2         3   Constant			      
0000240   2   1   0   1   0   2   2   2   1   0   1   0   1   0   2   2         4   Error			      
0000260   2   2   2   2   1   0   1   0   1   0   1   0   1   0   1   0         5   Builtin                           
0000300   2   2   1   0   1   0   2   2   1   0   1   0   2   2   2   1
0000320   0   1   0   1   0   2   2   1   0   1   0   2   2   1   0   1
0000340   0   2   2   2   1   0   1   0   1   0   2   2   1   0   1   0
0000360   2   2   1   0   1   0   2   1   0   2   1   0   2   1   0   2
0000400   1   0   2   1   0   2   1   0   2   1   0   2   1   0   2   1
0000420   0   2   1   0   2   1   0   2   1   0   2   1   0   2   1   0
0000440   2   1   0   2   1   0   2   1   0   2   1   0   2   1   0   2
0000460   1   0   2   1   0   2   1   0   2   1   0   2   1   0   2   1
0000500   0   2   1   0   0   1   1   0   1   0   1   0   1   0   2   2
0000520   2   2   0  24  63   0   2   1   0   2   2   0   8   0   5   0
0000540   2   1   0   2   2   2   0   6   0   5   0   4   0   2   0  24
0000560 130   1   0   1   0   1   0   2   2   2   2   0  24 108   2   2         24 xxx is the CBOR encoding of
0000600   2   0  18   2   0  24  84   1   0   0   1   2   0   4   0   3         a 1-byte integer xxx.  Integers
0000620   2   2   0  24  28   0   1   1   0   1   0   1   0   1   0   1         in [0..23] are represented 
0000640   0   1   0   1   0   1   0   1   0   0   4   1   0   2   2   0         directly as a single byte.
0000660   9   0   2   0   3   1   0   0  24 110   0  24 128   1   0   2
0000700   0  12   2   0  23   0   1   1   0   1   0   2   2   2   2   0
0000720  24 105   2   2   2   0  15   2   0  24  81   1   0   0   1   2
0000740   0  16   0   2   2   2   0  24  25   0   1   1   0   1   0   1
0000760   0   1   0   1   0   1   0   1   0   1   0   1   0   0   4   1
0001000   0   2   2   2   2   0  24 106   2   2   2   0  21   0  24  85
0001020   2   0   4   0   2   2   2   0  24  63   0   3   1   0   1   0
0001040   1   0   1   0   0   3   1   0   2   2   0   7   0   3   2   2
0001060   0  24  64   0   4   1   0   1   0   1   0   1   0   0   1   1
0001100   0   0  24 108   0  24 126   1   0   0  24 107   0  24 125   1
0001120   0   2   2   0  23   0   1   1   0   1   0   1   0   1   0   1
0001140   0   1   0   1   0   1   0   1   0   2   2   2   0  20   2   0
0001160  24 105   1   0   0   1   0  24 129   2   2   2   0  17   1   0
0001200   0   1   0  11   0   9   1   0   2   2   0  24  33   0   1   1
0001220   0   1   0   1   0   0   1   1   0   1   0   2   2   0  22   0
0001240   2   1   0   1   0   1   0   1   0   1   0   1   0   1   0   1
0001260   0   1   0   2   1   0   2   1   0   2   0   1   0   5   2   0
0001300   1   1   0   0   1   2   1   0   2   2   1   0   2   1   0   2
0001320   1   0   2   0   1   0   1   1   0   1   0   2   2   0   3   2
0001340   1   0   2   0   1   0   1   0   2   0   1   1   0   1   0   1
0001360   0   2   2   0   5   1   0   2   2   0   6   0   5   2   0   5
0001400   0   1   0   1   1   0   1   0   2   0   1   1   0   2   0   4
0001420   1   0   2   0   1   0   2   0   1   1   0   1   0   2   0   2
0001440   1   0   2   2   2   2   0  24 130   0   1   1   0   0  24 117
0001460   1   0   1   0   1   0   2   2   0  24 128   0   3   1   0   1
0001500   0   2   2   2   2   0  24 120   2   2   0  24 119   0  19   0
0001520   2   1   0   2   2   2   2   0  24 121   2   2   2   0  22   0
0001540  20   0  11   0   2   1   0   0  24 124   1   0   2   2   2   0
0001560  24 141   2   0  24  24   2   0  24  25   2   2   0  24 138   3     This section is a list of characters spelling
0001600   0  24 109   2   2   0  24 138   3   0  24  97   2   2   0  24
0001620 138   3   0  24 116   2   2   0  24 138   3   0  24  99   2   2        "matching pub key with invalid signature",
0001640   0  24 138   3   0  24 104   2   2   0  24 138   3   0  24 105
0001660   2   2   0  24 138   3   0  24 110   2   2   0  24 138   3   0     a Haskell string  from Ledger/Validation.hs
0001700  24 103   2   2   0  24 138   3   0  24  32   2   2   0  24 138     lifted into a Plutus Core value. The number
0001720   3   0  24 112   2   2   0  24 138   3   0  24 117   2   2   0     138 is the de Bruijn index of a variable 
0001740  24 138   3   0  24  98   2   2   0  24 138   3   0  24  32   2     pointing to `Cons`, a constructor for Scott-
0001760   2   0  24 138   3   0  24 107   2   2   0  24 138   3   0  24     encoded lists.  The string takes up something 
0002000 101   2   2   0  24 138   3   0  24 121   2   2   0  24 138   3     like 350 bytes ...
0002020   0  24  32   2   2   0  24 138   3   0  24 119   2   2   0  24
0002040 138   3   0  24 105   2   2   0  24 138   3   0  24 116   2   2
0002060   0  24 138   3   0  24 104   2   2   0  24 138   3   0  24  32
0002100   2   2   0  24 138   3   0  24 105   2   2   0  24 138   3   0
0002120  24 110   2   2   0  24 138   3   0  24 118   2   2   0  24 138
0002140   3   0  24  97   2   2   0  24 138   3   0  24 108   2   2   0
0002160  24 138   3   0  24 105   2   2   0  24 138   3   0  24 100   2
0002200   2   0  24 138   3   0  24  32   2   2   0  24 138   3   0  24     
0002220 115   2   2   0  24 138   3   0  24 105   2   2   0  24 138   3
0002240   0  24 103   2   2   0  24 138   3   0  24 110   2   2   0  24
0002260 138   3   0  24  97   2   2   0  24 138   3   0  24 116   2   2
0002300   0  24 138   3   0  24 117   2   2   0  24 138   3   0  24 114
0002320   2   2   0  24 138   3   0  24 101   0  24 139   1   0   2   0
0002340  10   0   7   0  24 142   0  24 141   1   0   2   0   8   0   5
0002360   0  24 140   0  24 135   1   0   1   0   1   0   2   1   0   2
0002400   2   0   1   0  24 105   0  24 104   2   2   2   5   0  15   0
0002420   3   0   2   0   1   1   0   2   1   0   0  24 120   2   5   1
0002440 101 116 114  97  99 101   0   1   2   0   1   1   0   0   1   2     "trace" (dynamic builtin name)
0002460   1   0   2   2   1   0   2   1   0   2   1   0   2   0   1   0
0002500   1   1   0   1   0   2   2   0   3   2   1   0   2   0   1   0
0002520   1   0   2   0   1   1   0   1   0   1   0   2   2   0   5   1
0002540   0   2   2   0   6   0   5   2   0   5   0   1   0   1   1   0
0002560   1   0   2   0   1   1   0   2   0   4   1   0   2   0   1   0
0002600   2   0   1   1   0   1   0   2   0   2   1   0   2   2   2   2
0002620   0  24 115   0   1   1   0   0  10   1   0   1   0   1   0   2
0002640   2   0  22   2   0  19   0   3   2   0   5   0   2   0  24 120
0002660   2   0   1   1   0   0   1   2   1   0   2   2   1   0   2   1
0002700   0   2   1   0   2   0   1   0   1   1   0   1   0   2   2   0
0002720   3   2   1   0   2   0   1   0   1   0   2   0   1   1   0   1
0002740   0   1   0   2   2   0   5   1   0   2   2   0   6   0   5   2
0002760   0   5   0   1   0   1   1   0   1   0   2   0   1   1   0   2
0003000   0   4   1   0   2   0   1   0   2   0   1   1   0   1   0   2
0003020   0   2   1   0   1   0   1   0   2   2   2   2   0  24 117   0
0003040   1   1   0   0  24 120   1   0   1   0   1   0   2   2   0  24
0003060 121   2   0   5   0   3   2   2   2   0   9   1   0   0   1   0
0003100   5   0   2   0  24 122   1   0   2   2   0  24  36   2   0   9
0003120   0   1   2   2   0  24  61   0  24  67   0  24  97   1   0   1
0003140   0   2   1   0   2   1   0   0   1   2   0   1   1   0   0   1
0003160   2   1   0   2   2   1   0   2   1   0   2   1   0   2   0   1
0003200   0   1   1   0   1   0   2   2   0   3   2   1   0   2   0   1
0003220   0   1   0   2   0   1   1   0   1   0   1   0   2   2   0   5
0003240   1   0   2   2   0   6   0   5   2   0   5   0   1   0   1   1
0003260   0   1   0   2   0   1   1   0   2   0   4   1   0   2   0   1
0003300   0   2   0   1   1   0   1   0   2   0   2   1   0   2   2   2
0003320   2   0  24 115   0   1   1   0   0   6   1   0   1   0   1   0
0003340   2   2   0   9   0   3   2   0   5   0   2   0  24 120   3   2
0003360  96   1   0   1   0   1   0   2   2   0  24  34   0   2   1   0
0003400   1   0   2   2   0  24  39   0   2   1   0   1   0   2   2   0
0003420  24  38   0   5   1   0   1   0   2   2   0  24  43   0   2   1
0003440   0   1   0   2   2   2   2   2   2   2   0  24  65   0  11   0
0003460   6   0   5   0   7   0   2   0   1   0   3   1   0   2   2   0
0003500   2   2   0   7   0   1   2   0   8   0   1   1   0   1   0   2
0003520   2   0  24  32   2   0   5   0   2   2   0   4   0   1   1   0
0003540   2   2   0  24  55   2   0  24  63   0   1   0  24  91   1   0
0003560   2   2   0  24  32   2   0  24  62   0   1   0  24  90   1   0
0003600   1   0   1   0   2   1   0   2   1   0   2   0   1   2   2   0
0003620  24  85   0   4   0   3   2   0   1   1   0   0   1   2   1   0
0003640   2   2   1   0   2   1   0   2   1   0   2   0   1   0   1   1
0003660   0   1   0   2   2   0   3   2   1   0   2   0   1   0   1   0
0003700   2   0   1   1   0   1   0   1   0   2   2   0   5   1   0   2
0003720   2   0   6   0   5   2   0   5   0   1   0   1   1   0   1   0
0003740   2   0   1   1   0   2   0   4   1   0   2   0   1   0   2   0
0003760   1   1   0   1   0   2   0   2   1   0   2   2   2   2   0  24
0004000 107   0   1   1   0   0  24  95   1   0   1   0   1   0   2   2
0004020   0  24 105   0   3   1   0   1   0   2   1   0   2   1   0   2
0004040   0   1   0   3   2   0   1   1   0   0   1   2   1   0   2   2
0004060   1   0   2   1   0   2   1   0   2   0   1   0   1   1   0   1
0004100   0   2   2   0   3   2   1   0   2   0   1   0   1   0   2   0
0004120   1   1   0   1   0   1   0   2   2   0   5   1   0   2   2   0
0004140   6   0   5   2   0   5   0   1   0   1   1   0   1   0   2   0
0004160   1   1   0   2   0   4   1   0   2   0   1   0   2   0   1   1
0004200   0   1   0   2   0   2   1   0   2   2   2   2   0  24 116   0
0004220   1   1   0   2   0  12   0   9   1   0   1   0   1   0   2   2
0004240   0  24 114   0   3   1   0   1   0   2   2   2   2   0  24 110
0004260   0   1   1   0   2   2   2   2   0  24 107   2   2   0  22   3
0004300   0   0   0   1   1   0   2   0   9   0   6   1   0   0  24 109
0004320   0  24 127   1   0   1   0   2   2   2   2   0  24 108   2   2
0004340   0  23   0   2   0   1   1   0   2   0  10   0   7   1   0   0
0004360  24 110   0  24 128   1   0   2   2   2   2   0  24 107   2   2
0004400   0  22   0   1   3   0   0   1   0   2   0   9   0   6   1   0
0004420   0  24 109   0  24 127   0  24 121   0  24 112   5   1 108  99
0004440 104  97 114  84 111  83 116 114 105 110 103   1   0   2   2   0      "charToString"  (dynamic builtin name)
0004460  24  41   0   1   1   0   1   0   1   0   1   0   0   4   1   0
0004500   2   2   0  24  40   0   1   1   0   1   0   1   0   1   0   0
0004520   2   5   1 102  97 112 112 101 110 100   1   0   1   0   1   0      "append"        (dynamic builtin name)
0004540   1   0   1   0   1   0   1   0   1   0   1   0   1   0   2   2
0004560   2   2   2   2   2   2   2   0   1   0  11   0  10   0   9   0
0004600   8   0   7   0   6   0   5   0   4   0   3   1   0   0   1   1
0004620   0   1   0   1   0   2   2   0   1   0   4   0   3   1   0   0
0004640   1   1   0   1   0   1   0   2   0   2   0   4   1   0   1   0
0004660   1   0   1   0   2   2   0   1   0   5   0   4   1   0   0   1
0004700   1   0   1   0   1   0   1   0   2   2   2   0   1   0   5   0
0004720   4   0   3   1   0   0   1   1   0   1   0   1   0   2   2   0
0004740   1   0   4   0   3   1   0   0   1   1   0   1   0   1   0   2
0004760   0   2   0   4   1   0   1   0   0   1   1   0   0   1   1   0
0005000   1   0   1   0   2   2   0   1   0   4   0   3   1   0   0   1
0005020   1   0   1   0   1   0   2   2   0   1   0   4   0   3   1   0
0005040   0   1   1   0   1   0   1   0   1   0   1   0   1   0   2   0
0005060   5   0   7   1   0   1   0   1   0   1   0   1   0   1   0   1
0005100   0   2   2   0   4   0   8   0   7   1   0   1   0   1   0   1
0005120   0   1   0   1   0   2   0   3   0   7   1   0   1   0   1   0
0005140   1   0   1   0   1   0   2   0   2   0   7   1   0   1   0   1
0005160   0   1   0   1   0   1   0   2   0   1   0   7   1   0   0   1
0005200   1   0   1   0   0   2   1   0   1   0   0   1   1   0   0   1
0005220   1   0   1   0   1   0   1   0   1   0   2   2   2   2   0   1
0005240   0   6   0   5   0   4   0   3   1   0   0   1   1   0   1   0
0005260   1   0   1   0   2   2   2   0   1   0   5   0   4   0   3   1
0005300   0   0   1   1   0   1   0   1   0   1   0   1   0   1   0   1
0005320   0   2   2   2   2   0  24  45   2   2   2   2   2   0  13   0
0005340   7   0   6   0   5   0   3   0   2   1   0   2   2   2   0  10
0005360   0   8   0   2   0   5   1   0   0  24  47   0  24  65   1   0
0005400   1   0   1   0   2   2   0   5   0   2   1   0   1   0   2   2
0005420   0   7   0   3   1   0   1   0   2   2   2   2   2   0  13   0
0005440   7   0   4   0   3   0   2   0   1   1   0   1   0   1   0   2
0005460   2   0   1   0   4   0   3   1   0   0   1   1   0   1   0   1
0005500   0   1   0   1   0   2   1   0   2   2   2   2   2   0  10   0
0005520   5   1   0   1   0   2   1   0   2   2   2   2   2   0  13   0
0005540   6   1   0   1   0   2   2   2   2   2   0  15   0   6   1   0
0005560   1   0   2   2   2   2   2   0  17   0   5   1   0   1   0   2
0005600   2   2   2   2   0  24  36   2   2   2   0  18   0  15   0   4
0005620   0   2   1   0   2   2   2   2   0  24  49   0  12   1   0   0
0005640  15   1   0   0  24  52   0  24  69   1   0   0  24  50   1   0
0005660   0  24  51   0  24  68   1   0   2   1   0   4   0  24  67   1
0005700   0   0  24  49   0  24  66   1   0   2   1   0   4   0  24  65
0005720   1   0   2   2   2   2   2   0  16   0   4   1   0   1   0   0
0005740  24  48   1   0   2   1   0   4   0  24  66   1   0   2   2   2
0005760   2   0  24  46   0   9   1   0   0  12   1   0   0  24  49   0
0006000  24  66   0  24  65   0  24  64   1   0   0  24  44   1   0   2
0006020   2   2   2   2   0  14   0   5   1   0   1   0   2   2   2   2
0006040   2   0  16   0   4   1   0   1   0   2   2   2   2   2   0  24
0006060  35   2   2   2   0  17   0  14   0   4   0   2   1   0   2   2
0006100   2   2   0  24  48   0  11   1   0   0  14   1   0   0  24  51
0006120   0  24  68   1   0   0  24  49   1   0   0  24  50   0  24  67
0006140   1   0   2   1   0   4   0  24  66   1   0   0  24  48   0  24
0006160  65   1   0   2   1   0   4   0  24  64   1   0   2   2   2   2
0006200   2   0  15   0   3   1   0   1   0   0  24  47   1   0   2   1
0006220   0   4   0  24  65   1   0   2   2   2   2   0  24  45   0   8
0006240   1   0   0  11   1   0   0  24  48   0  24  65   0  24  64   0
0006260  24  63   0  24  62   0   5   1   0   2   2   2   2   2   0  11
0006300   0   4   1   0   1   0   0  24  44   1   0   2   2   2   2   0
0006320  24  41   0   4   1   0   0   7   1   0   0  24  44   0  24  61
0006340   1   0   0  24  43   0  24  60   1   0   2   1   0   2   2   2
0006360   2   2   0  12   0   5   1   0   1   0   2   2   2   2   2   0
0006400  14   0   5   1   0   1   0   2   2   2   2   2   0  16   0   5
0006420   1   0   1   0   2   2   2   2   2   0  24  35   2   2   2   0
0006440  17   0  14   0   4   0   2   1   0   2   2   2   2   0  24  48
0006460   0  11   1   0   0  14   1   0   0  24  51   0  24  68   1   0
0006500   0  24  49   1   0   0  24  50   0  24  67   1   0   2   1   0
0006520   4   0  24  66   1   0   0  24  48   0  24  65   1   0   2   1
0006540   0   4   0  24  64   1   0   2   2   2   2   2   0  15   0   4
0006560   1   0   1   0   0  24  47   1   0   2   1   0   4   0  24  65
0006600   1   0   2   2   2   2   0  24  45   0   8   1   0   0  11   1
0006620   0   0  24  48   0  24  65   0  24  64   0  24  63   1   0   0
0006640  24  43   1   0   2   2   2   2   2   0  13   0   4   1   0   1
0006660   0   2   2   2   2   2   0  15   0   4   1   0   1   0   2   2
0006700   2   2   2   0  24  34   2   2   2   0  16   0  13   0   4   0
0006720   2   1   0   2   2   2   2   0  24  47   0  10   1   0   0  13
0006740   1   0   0  24  50   0  24  67   1   0   0  24  48   1   0   0
0006760  24  49   0  24  66   1   0   2   1   0   4   0  24  65   1   0
0007000   0  24  47   0  24  64   1   0   2   1   0   4   0  24  63   1
0007020   0   2   2   2   2   2   0  14   0   3   1   0   1   0   0  24
0007040  46   1   0   2   1   0   4   0  24  64   1   0   2   2   2   2
0007060   0  24  44   0   7   1   0   0  10   1   0   0  24  47   0  24
0007100  64   0  24  63   0  24  62   0  24  61   0   4   0  24  59   0
0007120   4   1   0   1   0   1   0   1   0   1   0   2   1   0   2   2
0007140   2   2   2   0   9   0   5   1   0   1   0   2   1   0   2   2
0007160   2   2   2   0  12   0   6   1   0   1   0   2   2   2   2   2
0007200   0  14   0   6   1   0   1   0   2   2   2   2   2   0  16   0
0007220   5   1   0   1   0   2   2   2   2   2   0  24  35   2   2   2
0007240   0  17   0  15   0   4   0   2   1   0   2   2   2   2   0  24
0007260  48   0  14   1   0   0  13   1   0   0  24  51   0  24  68   1
0007300   0   0  24  49   1   0   0  24  50   0  24  67   1   0   2   1
0007320   0   4   0  24  66   1   0   0  24  48   0  24  65   1   0   2
0007340   1   0   4   0  24  64   1   0   2   2   2   2   2   0  15   0
0007360   4   1   0   1   0   0  24  47   1   0   2   1   0   4   0  24
0007400  65   1   0   2   2   2   2   0  24  45   0  11   1   0   0  10
0007420   1   0   0  24  48   0  24  65   0  24  64   0  24  63   1   0
0007440   0  24  43   1   0   2   2   2   2   2   0  13   0   5   1   0
0007460   1   0   2   2   2   2   2   0  15   0   4   1   0   1   0   2
0007500   2   2   2   2   0  24  34   2   2   2   0  16   0  14   0   4
0007520   0   2   1   0   2   2   2   2   0  24  47   0  13   1   0   0
0007540  12   1   0   0  24  50   0  24  67   1   0   0  24  48   1   0
0007560   0  24  49   0  24  66   1   0   2   1   0   4   0  24  65   1
0007600   0   0  24  47   0  24  64   1   0   2   1   0   4   0  24  63
0007620   1   0   2   2   2   2   2   0  14   0   3   1   0   1   0   0
0007640  24  46   1   0   2   1   0   4   0  24  64   1   0   2   2   2
0007660   2   0  24  44   0  10   1   0   0   9   1   0   0  24  47   0
0007700  24  64   0  24  63   0  24  62   0  24  61   0   5   1   0   2
0007720   2   2   2   2   0  10   0   4   1   0   1   0   0  24  43   1
0007740   0   2   2   2   2   0  24  40   0   6   1   0   0   5   1   0
0007760   0  24  43   0  24  60   1   0   0  24  42   0  24  59   1   0
0010000   2   1   0   2   2   2   2   2   0  11   0   5   1   0   1   0
0010020   2   2   2   2   2   0  13   0   5   1   0   1   0   2   2   2
0010040   2   2   0  15   0   5   1   0   1   0   2   2   2   2   2   0
0010060  24  34   2   2   2   0  16   0  14   0   4   0   2   1   0   2
0010100   2   2   2   0  24  47   0  13   1   0   0  12   1   0   0  24
0010120  50   0  24  67   1   0   0  24  48   1   0   0  24  49   0  24
0010140  66   1   0   2   1   0   4   0  24  65   1   0   0  24  47   0
0010160  24  64   1   0   2   1   0   4   0  24  63   1   0   2   2   2
0010200   2   2   0  14   0   4   1   0   1   0   0  24  46   1   0   2
0010220   1   0   4   0  24  64   1   0   2   2   2   2   0  24  44   0
0010240  10   1   0   0   9   1   0   0  24  47   0  24  64   0  24  63
0010260   0  24  62   1   0   0  24  42   1   0   2   2   2   2   2   0
0010300  12   0   4   1   0   1   0   2   2   2   2   2   0  14   0   4
0010320   1   0   1   0   2   2   2   2   2   0  24  33   2   2   2   0
0010340  15   0  13   0   4   0   2   1   0   2   2   2   2   0  24  46
0010360   0  12   1   0   0  11   1   0   0  24  49   0  24  66   1   0
0010400   0  24  47   1   0   0  24  48   0  24  65   1   0   2   1   0
0010420   4   0  24  64   1   0   0  24  46   0  24  63   1   0   2   1
0010440   0   4   0  24  62   1   0   2   2   2   2   2   0  13   0   3
0010460   1   0   1   0   0  24  45   1   0   2   1   0   4   0  24  63
0010500   1   0   2   2   2   2   0  24  43   0   9   1   0   0   8   1
0010520   0   0  24  46   0  24  63   0  24  62   0  24  61   0  24  60
0010540   0   4   0  24  58   0   4   1   0   2   2   0  12   0   1   1
0010560   0   1   0   1   0   1   0   1   0   1   0   1   0   1   0   0
0010600   7   1   0   1   0   1   0   1   0   2   0   3   0   5   1   0
0010620   1   0   1   0   0   2   1   0   1   0   1   0   0   1   1   0
0010640   0   1   1   0   2   2   2   2   2   2   2   2   0   6   0  12
0010660   0  10   0   2   0  11   0   3   0   4   0   9   0   8   1   0
0010700   1   0   2   1   0   2   2   0   1   0  24  28   0  24  27   2
0010720   2   5   0   5   0   2   0   1   1   0   1   0   2   1   0   2
0010740   2   0   1   0  24  27   0  24  26   2   2   5   0   7   0   2
0010760   0   1   1   0   1   0   2   1   0   2   2   0   1   0  24  26
0011000   0  24  25   2   2   5   0   8   0   2   0   1   1   0   1   0
0011020   1   0   1   0   1   0   1   0   1   0   1   0   1   0   2   2
0011040   2   2   2   2   2   2   0   1   0  10   0   9   0   8   0   7
0011060   0   6   0   5   0   4   0   3   1   0   0   1   1   0   1   0
0011100   2   2   2   2   0  19   2   2   0   5   0   2   0   1   1   0
0011120   0   3   1   0   0   2   0  24  39   1   0   1   0   2   2   2
0011140   2   0  18   2   2   0   4   0   2   0   1   1   0   0   2   1
0011160   0   0   3   0  24  38   1   0   1   0   2   2   2   2   0  17
0011200   2   2   0   4   0   2   0   1   1   0   0   9   1   0   2   2
0011220   2   2   0  18   2   2   0   4   0   3   0   2   1   0   0   8
0011240   1   0   0   9   0  24  38   0  24  37   1   0   1   0   2   1
0011260   0   2   2   0   1   0  19   0  18   2   2   5   0   6   0   2
0011300   0   1   1   0   1   0   2   1   0   2   2   0   1   0  18   0
0011320  17   2   2   5   0   9   0   2   0   1   1   0   1   0   1   0
0011340   0   3   1   0   1   0   1   0   0   2   1   0   1   0   1   0
0011360   0   1   1   0   0   1   1   0   2   0   2   0  17   1   0   1
0011400   0   1   0   2   1   0   2   1   0   2   0   1   2   2   0   6
0011420   0   4   0   3   2   0   1   1   0   0   1   2   1   0   2   2
0011440   1   0   2   1   0   2   1   0   2   0   1   0   1   1   0   1
0011460   0   2   2   0   3   2   1   0   2   0   1   0   1   0   2   0
0011500   1   1   0   1   0   1   0   2   2   0   5   1   0   2   2   0
0011520   6   0   5   2   0   5   0   1   0   1   1   0   1   0   2   0
0011540   1   1   0   2   0   4   1   0   2   0   1   0   2   0   1   1
0011560   0   1   0   2   0   2   1   0   2   2   2   2   0  24  28   0
0011600   1   1   0   0  24  31   1   0   1   0   1   0   2   2   0  24
0011620  26   0   3   1   0   1   0   2   2   0  24  34   2   2   0  24
0011640  29   0   2   2   1   0   2   1   0   2   0   1   0   3   2   0
0011660   1   1   0   0   1   2   1   0   2   2   1   0   2   1   0   2
0011700   1   0   2   0   1   0   1   1   0   1   0   2   2   0   3   2
0011720   1   0   2   0   1   0   1   0   2   0   1   1   0   1   0   1
0011740   0   2   2   0   5   1   0   2   2   0   6   0   5   2   0   5
0011760   0   1   0   1   1   0   1   0   2   0   1   1   0   2   0   4
0012000   1   0   2   0   1   0   2   0   1   1   0   1   0   2   0   2
0012020   1   0   2   2   2   2   0  24  37   0   1   1   0   0  24  40
0012040   1   0   1   0   1   0   2   2   0  24  35   0   3   1   0   1
0012060   0   2   2   0  24  43   2   2   0  24  38   0   2   2   2   2
0012100   2   0  24  31   0   1   1   0   2   2   0  22   3   0   0   0
0012120   1   1   0   1   0   2   2   0  23   0   2   0   1   1   0   2
0012140   2   0  22   0   1   3   0   0   2   0   7   0   4   0  24  42
0012160   2   0   7   0   4   0  24  33   1   0   1   0   2   1   0   2
0012200   1   0   2   0   1   2   2   2   0   5   0   8   0   4   0   3
0012220   2   0   1   1   0   0   1   2   1   0   2   2   1   0   2   1
0012240   0   2   1   0   2   0   1   0   1   1   0   1   0   2   2   0
0012260   3   2   1   0   2   0   1   0   1   0   2   0   1   1   0   1
0012300   0   1   0   2   2   0   5   1   0   2   2   0   6   0   5   2
0012320   0   5   0   1   0   1   1   0   1   0   2   0   1   1   0   2
0012340   0   4   1   0   2   0   1   0   2   0   1   1   0   1   0   2
0012360   0   2   1   0   2   2   2   2   0  24  26   0   1   1   0   0
0012400  24  29   1   0   1   0   1   0   2   2   0  24  24   0   3   1
0012420   0   1   0   2   2   0  24  32   2   2   0  24  27   0   2   2
0012440   2   2   2   0  20   0   1   1   0   2   1   0   2   1   0   2
0012460   0   1   0   3   2   0   1   1   0   0   1   2   1   0   2   2
0012500   1   0   2   1   0   2   1   0   2   0   1   0   1   1   0   1
0012520   0   2   2   0   3   2   1   0   2   0   1   0   1   0   2   0
0012540   1   1   0   1   0   1   0   2   2   0   5   1   0   2   2   0
0012560   6   0   5   2   0   5   0   1   0   1   1   0   1   0   2   0
0012600   1   1   0   2   0   4   1   0   2   0   1   0   2   0   1   1
0012620   0   1   0   2   0   2   1   0   2   2   2   2   0  24  36   0
0012640   1   1   0   0  24  39   1   0   1   0   1   0   2   2   0  24
0012660  34   0   3   1   0   1   0   2   2   0  24  42   2   2   0  24
0012700  37   0   2   2   0  24  33   0   1   2   0   7   0   4   0  24
0012720  41   1   0   1   0   2   2   2   0  14   0  17   0   2   0   1
0012740   1   0   2   1   0   2   1   0   2   0   1   0   3   2   0   1
0012760   1   0   0   1   2   1   0   2   2   1   0   2   1   0   2   1
0013000   0   2   0   1   0   1   1   0   1   0   2   2   0   3   2   1
0013020   0   2   0   1   0   1   0   2   0   1   1   0   1   0   1   0
0013040   2   2   0   5   1   0   2   2   0   6   0   5   2   0   5   0
0013060   1   0   1   1   0   1   0   2   0   1   1   0   2   0   4   1
0013100   0   2   0   1   0   2   0   1   1   0   1   0   2   0   2   1
0013120   0   2   2   2   2   0  24  36   0   1   1   0   0  24  39   1
0013140   0   1   0   1   0   2   2   0  24  34   0   3   1   0   1   0
0013160   2   2   0  24  42   2   2   0  24  37   0   2   2   0  24  31
0013200   0   1   2   0   7   0   4   0  24  41   2   0   7   0   4   0
0013220  24  31   1   0   1   0   1   0   2   2   2   2   0   7   1   0
0013240   0   1   0  24  26   2   2   2   0  23   1   0   0   1   1   0
0013260   2   2   0  21   0   1   1   0   1   0   2   2   0  24  24   0
0013300   2   2   0  20   0   1   2   2   2   2   0   7   1   0   0   1
0013320   1   0   1   0   2   1   0   2   2   0  23   0   3   1   0   1
0013340   0   2   2   2   2   0  15   2   2   2   2   0  12   1   0   0
0013360   1   1   0   1   0   2   2   2   2   0  17   0   1   1   0   0
0013400  20   1   0   2   2   0  24  28   0   3   1   0   1   0   2   2
0013420   0  13   0   2   0   7   0  24  37   0  16   0   7   1   0   0
0013440   5   1   0   2   2   0  24  32   0   4   0   5   0  24  35   0
0013460   2   0  24  27   0   1   2   2   2   0  23   1   0   0   1   1
0013500   0   2   2   0  21   0   1   1   0   1   0   2   2   0  24  24
0013520   0   2   2   1   0   2   1   0   2   0   1   0   6   2   0   1
0013540   1   0   0   1   2   1   0   2   2   1   0   2   1   0   2   1
0013560   0   2   0   1   0   1   1   0   1   0   2   2   0   3   2   1
0013600   0   2   0   1   0   1   0   2   0   1   1   0   1   0   1   0
0013620   2   2   0   5   1   0   2   2   0   6   0   5   2   0   5   0
0013640   1   0   1   1   0   1   0   2   0   1   1   0   2   0   4   1
0013660   0   2   0   1   0   2   0   1   1   0   1   0   2   0   2   1
0013700   0   2   2   2   2   0  24  32   0   1   1   0   2   0  23   0
0013720   6   1   0   1   0   1   0   2   2   0  24  30   0   3   1   0
0013740   1   0   2   2   2   2   0  22   2   2   0  15   0   2   0  11
0013760   1   0   2   2   0  24  29   0  11   0   2   1   0   2   0   8
0014000   0   5   0  24  42   0  24  37   0   2   2   0   1   1   0   0
0014020   1   2   1   0   2   2   1   0   2   1   0   2   1   0   2   0
0014040   1   0   1   1   0   1   0   2   2   0   3   2   1   0   2   0
0014060   1   0   1   0   2   0   1   1   0   1   0   1   0   2   2   0
0014100   5   1   0   2   2   0   6   0   5   2   0   5   0   1   0   1
0014120   1   0   1   0   2   0   1   1   0   2   0   4   1   0   2   0
0014140   1   0   2   0   1   1   0   1   0   2   0   2   1   0   1   0
0014160   1   0   1   0   2   2   2   2   0  24  26   0   1   1   0   0
0014200   3   1   0   1   0   1   0   2   2   0   6   0   3   2   2   2
0014220   2   0  10   1   0   0   1   0   6   0   5   0   2   0  24  31
0014240   1   0   1   0   2   1   0   2   2   0   1   0   6   0   5   2
0014260   2   5   0  16   0   2   0   1   1   0   1   0   0   2   1   0
0014300   1   0   0   1   1   0   0   1   1   0   1   0   1   0   1   0
0014320   2   0   3   0   5   1   0   1   0   1   0   1   0   1   0   2
0014340   2   0   2   0   6   0   5   1   0   1   0   1   0   1   0   2
0014360   0   1   0   5   1   0   0   1   5   0   0   1   0   1   0   1
0014400   0   2   2   0   1   0   4   0   3   1   0   0   1   2   0   1
0014420   1   0   0   1   2   1   0   2   2   1   0   2   1   0   2   1
0014440   0   2   0   1   0   1   1   0   1   0   2   2   0   3   2   1
0014460   0   2   0   1   0   1   0   2   0   1   1   0   1   0   1   0
0014500   2   2   0   5   1   0   2   2   0   6   0   5   2   0   5   0
0014520   1   0   1   1   0   1   0   2   0   1   1   0   2   0   4   1
0014540   0   2   0   1   0   2   0   1   1   0   1   0   2   0   2   1
0014560   0   1   0   1   0   2   2   2   2   0   9   0   1   1   0   0
0014600  12   1   0   1   0   1   0   2   2   0  13   2   0   5   0   3
0014620   2   2   2   0   9   1   0   0   1   0   5   0   2   0  14   1
0014640   0   1   0   0   2   1   0   1   0   1   0   1   0   2   2   0
0014660   1   0   5   0   4   1   0   0   1   1   0   0   1   1   0   0
0014700   1
```
